program recover;

{$APPTYPE CONSOLE}

uses
  SysUtils, Classes;

type
  EGetOptError = class(Exception);

var
  Options: record
    BlockSize: Integer;
    InputFile: String;
    FormatStr: String;
    Index: Integer;
  end;

procedure CopyJpegFrom(Src, Dst: TStream);
var
  Buffer: PByteArray;
  Size: Integer;
begin
  GetMem(Buffer,$10000);
  try
    while TRUE do
    begin
      FillChar(Buffer^,4,#0);
      Src.Read(Buffer^,4);
      Size:= ((Buffer[2] shl 8) + Buffer[3]) - 2;
      case Buffer[1] of
        $01,$D0..$D8: begin
                        Dst.WriteBuffer(Buffer^,2);
                        Src.Seek(-2,soFromCurrent);
                      end;
        $D9: begin
               Dst.WriteBuffer(Buffer^,2);
               // Salimos del procedure
               Exit;
             end;
        $DA: begin
               Dst.WriteBuffer(Buffer^,2);
               Src.Seek(-2,soFromCurrent);
               while TRUE do
               begin
                 Src.ReadBuffer(Buffer^,1);
                 if Buffer[0] = $FF then
                 begin
                   Src.ReadBuffer(Buffer^,1);
                  if (Buffer[0] <> 0) and not (Buffer[0] in [$D0..$D7]) then
                   begin
                     Src.Seek(-2,soFromCurrent);
                     break;
                   end else
                   begin
                     Buffer[1]:= Buffer[0];
                     Buffer[0]:= $FF;
                     Dst.WriteBuffer(Buffer^,2);
                   end;
                 end else
                   Dst.WriteBuffer(Buffer^,1);
               end;
             end
        else
        begin
          Dst.WriteBuffer(Buffer^,4);
          Src.ReadBuffer(Buffer^,Size);
          Dst.WriteBuffer(Buffer^,Size);
        end;
      end;
    end;
  finally
    FreeMem(Buffer);
  end;
end;

procedure SaveJpegToFile(Src: TStream; Filename: String);
var
  Stream: TMemoryStream;
begin
  Stream:= TMemoryStream.Create;
  try
    CopyJpegFrom(Src, Stream);
    Stream.SaveToFile(Filename);
  finally
    Stream.Free;
  end;
end;

procedure ScanFile(Stream: TStream; FormatStr: String; Index: Integer); overload;
var
  B: Byte;
  Leidos: Integer;
  Posicion: int64;
  Str: String;
begin
  // Vamos leyendo byte a byte el archivo
  Leidos:= Stream.Read(B,1);
  while Leidos = 1 do
    if B = $FF then
    begin
      Leidos:= Stream.Read(B,1);
      if Leidos = 1 then
        if B = $D8 then
        begin
          Leidos:= Stream.Read(B,1);
          if Leidos = 1 then
            if B = $FF then
            begin
              // Hasta encontrar la combinacion $FFD8FF
              Writeln('Encontrada una posible imagen');
              // Guardamos la posicion
              Posicion:= Stream.Position;
              Stream.Seek(-3,soFromCurrent);
              Str:= Format(FormatStr,[index]);
              try
                SaveJpegToFile(Stream,Str);
                inc(index);
                Writeln('Recuperada: ' + Str);
              except
                // Si se produce un error lo mostramos, pero no salimos del bucle
                On E: Exception do
                   Writeln('Error: ' + E.Message);
              end;
              // Volvemos a la posicion en la que estabamos
              Stream.Position:= Posicion;
            end;
        end;
    end else
      Leidos:= Stream.Read(B,1);
end;

const
  MAXSIZE = 8*1024*1024;

procedure ScanFile(Stream: TStream; FormatStr: String; Index: Integer;
  BlockSize: Integer); overload;
var
  Block: PChar;
  Leidos: Integer;
  Posicion: int64;
  Str: String;
  TmpStream: TMemorystream;
begin
  GetMem(Block,BlockSize);
  try
    // Vamos leyendo bloque a bloque el archivo
    Leidos:= Stream.Read(Block^,BlockSize);
    while Leidos = BlockSize do
    begin
      // Hasta encontrar la combinacion $FFD8FF
      if CompareMem(Block,PChar(#$FF#$D8#$FF),3) then
      begin
        Posicion:= Stream.Position;
        TmpStream:= TMemoryStream.Create;
        try
          TmpStream.Write(Block^,Leidos);
          TmpStream.CopyFrom(Stream,MAXSIZE - (MAXSIZE mod BlockSize));
          TmpStream.Position:= 0;
          Writeln('Encontrada una posible imagen.');
          Str:= Format(FormatStr,[index]);
          // Intentamos recuperar la imagen
          try
            SaveJpegToFile(TmpStream,Str);
            inc(index);
            Writeln('Recuperada: ' + Str);
          except
            // Si se produce un error lo mostramos, pero no salimos del bucle
            On E: Exception do
               Writeln('Error: ' + E.Message);
          end;
        finally
          TmpStream.Free;
        end;
        // Nos movemos hasta el siguiente cluster
        Stream.Position:= Posicion;
      end;
      Leidos:= Stream.Read(Block^,BlockSize);
    end;
  finally
    FreeMem(Block);
  end;
end;

procedure WriteHelp;
begin
  Writeln;
  Writeln('Uso: recover [opciones]');
  Writeln('  --bs');
  Writeln('    Medida del bloque en el archivo de entrada');
  Writeln('  --fs');
  Writeln('    Cadena para dar formato a las imagenes recuperadas.');
  Writeln('  --help -h');
  Writeln('    Muestra este mensaje de ayuda');
  Writeln('  --id');
  Writeln('    Valor inicial del indice');
  Writeln('  --if');
  Writeln('    Nombre del archivo o disco a explorar'); 
  Writeln;
  Writeln('Ejemplos:');
  Writeln('  recover --if=\\.\F: --bs=4096 --fs=%d.jpg --id=5');
  Writeln('    Busca y repuepar imagenes del disco F, y las guarda');
  Writeln('    como 5.jpg, 6.jpg, ...');
  Writeln('  recover --if=thumb.db --fs=%d.jpg');
  Writeln('    Busca y repuepar imagenes en el archivo "thumb.db", y');
  Writeln('    las guarda como 0.jpg, 1.jpg, ...');
  Writeln;      
end;

function Starts(Sub, Str: String): Boolean;
begin
  Result:= SameText(Sub,Copy(Str,1,Length(sub)));
end;

procedure Getopt;
var
  i: Integer;
  Str: String;
begin
  FillChar(Options,Sizeof(Options),#0);
  if ParamCount > 0 then
    for i:= 1 to ParamCount do
    begin
      Str:= ParamStr(i);
      if (Copy(Str,1,2) = '--') and (Length(Str) > 2) then
      begin
        if Starts('--if=',Str) then
        begin
          if Options.InputFile = EmptyStr then
            Options.InputFile:= Copy(Str,Length('--if=') + 1,MAXINT)
          else
            raise EGetOptError.Create(Str);
        end else if Starts('--id=',Str) then
        begin
          if Options.Index = 0 then
            Options.Index:= StrToInt(Copy(Str,Length('--id=') + 1,MAXINT))
          else
            raise EGetOptError.Create(Str);
        end else if Starts('--fs=',Str) then
        begin
          if Options.FormatStr = EmptyStr then
            Options.FormatStr:= Copy(Str,Length('--fs=') + 1,MAXINT)
          else
            raise EGetOptError.Create(Str);
        end else if Starts('--bs=',Str) then
        begin
          if Options.BlockSize = 0 then
            Options.BlockSize:= StrToInt(Copy(Str,Length('--bs=') + 1,MAXINT))
          else
            raise EGetOptError.Create(Str);
        end else if SameText('--help',Str) then
        begin
          WriteHelp;
          Halt;
        end else
          raise EGetOptError.Create(Str);
      end else if (Copy(Str,1,1) = '-') and (Length(Str) > 1) then
      begin
        if SameText('-h',Str) then
        begin
          WriteHelp;
          Halt;
        end else
          raise EGetOptError.Create(Str);
      end else
      begin
        raise EGetOptError.Create(Str);
      end;
    end else
    begin
      WriteHelp;
      Halt;
    end;
end;

procedure WriteHeader;
begin
  Writeln;
  Writeln('Recover 0.2 - Copyright (c) 2007 Domingo Seoane');
  Writeln('http://delphi.jmrds.com');
  Writeln;
end;   

var
  Stream: TFileStream;
begin
  WriteHeader;
  try
    GetOpt;
    TFileStream.Create(Format(Options.FormatStr,[0]),fmCreate).Free;
    DeleteFile(Format(Options.FormatStr,[0]));
    Stream:= TFileStream.Create(Options.InputFile,fmOpenRead or
      fmShareDenyNone);
    try
      Writeln('Explorando el archivo: ' + Options.InputFile);
      if Options.BlockSize > 0 then
        ScanFile(Stream,Options.FormatStr,Options.Index,Options.BlockSize)
      else
        ScanFile(Stream,Options.FormatStr,Options.Index);
    finally
      Stream.Free;
    end;
  except
    // Si se produce un error, lo mostramos
    On E: Exception do
      Writeln('Error: ' + E.Message);
  end;
end.
